#!/usr/bin/env python

import md5
import sys
import signal
import os

from socket import socket, gethostbyname, SOCK_STREAM, AF_INET, error
from select import select
from getopt import getopt
from string import atoi, split, strip
from random import randint
from sys import exit, argv
from estutil import true, false, striplist, dict_pkr, packdict, dict_pkra, unpackdict, dict_ske, localhost, dict_port, sigchld_handler, ctrlc_handler
from os import fork, execv
from time import time, strftime, localtime

try:
    from syslog import syslog
    syslog_support = true
except:
    print "Warning: your version of python doesn't have syslog support compiled in !"
    syslog_support = false

from Crypto.Cipher import ARC4
from Crypto.Util import randpool
from CryptoWorld.PublicKey import RSA

class Parameter:

    def __init__(self):
	self.mappings = None
	self.verbose = false
	self.syslog = false

    def usage(self):
	print "Usage: est.py -m <mappings> -s -v"
	exit()

    def parse(self):
	try:
	    optlist, args = getopt(argv[1:], "m:sv")
	except:
	    self.usage()

	for flag, arg in optlist:
	    if (flag == "-m"): self.mappings = arg
	    elif (flag == "-v"): self.verbose = true
	    elif (flag == "-s") and (syslog_support == true): self.syslog = true

	if (len(args) > 0) or (self.mappings == None):
	    self.usage()

class Client:

    def __init__(self):
	self.publickey = {}
	self.portlist = {}
	self.clientcrypto = {}                             # Cipher for stuff the client encrypts
	self.daemoncrypto = {}                             # Cipher for stuff to decrypt from the daemon
	self.verbose = false
	self.syslog = None

    def log(self, message):
	if (self.syslog == true):
	    syslog("est.py: %s" % message)
	if (self.verbose == true):
	    sys.stdout.write("%s est.py: %s\n" % (strftime("%b %d %X", localtime(time())), message))

    def initialize(self, mappings, verbose, syslog):

	self.verbose = verbose
	self.syslog = syslog

	self.log("Initializing ECLiPt Secure Tunnel Client")

	pool = randpool.RandomPool()

	if not os.path.exists(mappings):
	    self.log("Mappings file %s doesn't exist" % mappings)
	    exit(1)

	file = open(mappings, "r")
	maplist = striplist(file.readlines())
	file.close()

	for i in range(len(maplist)):

	    if (maplist[i][0] != "#") and (maplist[i] != ""):
		splitted = split(maplist[i], ":")

		if (len(splitted) != 4):
		    self.log("Invalid entry in %s line %d" % (mappings, i))
		else:

		    localport = atoi(strip(splitted[0]))
		    desthost = gethostbyname(strip(splitted[1]))
		    destport = atoi(strip(splitted[2]))
		    destforward = atoi(strip(splitted[3]))

		    self.portlist[localport] = (desthost, destport, destforward)

		    if not self.publickey.has_key(desthost):   # Request public key

			self.log("Requesting public key from %s:%d" % (desthost, destport))
			server = socket(AF_INET, SOCK_STREAM)
			try:
			    server.connect(desthost, destport)
			except error, detail:
			    self.log("Couldn't connect to %s:%d: %s" % (desthost, destport, detail[1]))
			    exit(1)
			server.send(packdict({dict_pkr:""}))
			length = atoi(server.recv(4))
			self.publickey[desthost] = RSA()
			self.publickey[desthost].set_public_key(unpackdict(server.recv(length))[dict_pkra])
			server.close()
			self.log("Got public key from %s:%d" % (desthost, destport))

		    if not self.clientcrypto.has_key(desthost):# There's no session key yet

			self.log("Sending encrypted session key to %s:%d" % (desthost, destport))
			sessionkey = md5.new(pool.getBytes(64)).digest()
			
			sessionkey_encrypted = self.publickey[desthost].public_encrypt(sessionkey)

			self.clientcrypto[desthost] = ARC4.new(sessionkey)
			self.daemoncrypto[desthost] = ARC4.new(sessionkey)

			server = socket(AF_INET, SOCK_STREAM)
			server.connect(desthost, destport)
			server.send(packdict({dict_ske:sessionkey_encrypted}))
			server.close()
			self.log("Encrypted session key sent to %s:%d" % (desthost, destport))

    def listen(self):

	sockets = []
	for i in range(len(self.portlist.keys())):
	    listensocket = socket(AF_INET, SOCK_STREAM)
	    try:
		listensocket.bind(localhost, self.portlist.keys()[i])
	    except error, detail:
		self.log("[ port %d ] Bind failed: %s" % (self.portlist.keys()[i], detail[1]))
		exit(1)
	    listensocket.listen(2)
	    sockets.append(listensocket)

	prettyportlist = ""
	for i in range(len(self.portlist.keys()) - 1):
	    prettyportlist = "%s%d, " % (prettyportlist, self.portlist.keys()[i])
	prettyportlist = "%s%d" % (prettyportlist, self.portlist.keys()[len(self.portlist.keys()) - 1])

	if len(self.portlist.keys()) > 1:
	    self.log("[ ports %s ] Listening" % prettyportlist)
	else:
	    self.log("[ port %s ] Listening" % prettyportlist)

	while 1:
	    try:
		[iwtd, owtd, ewtd] = select (sockets, [], [])
	    except:
		continue

	    for in_sock in iwtd:
		in_conn, in_addr = in_sock.accept()
		localport = in_sock.getsockname()[1]

		destination = self.portlist[localport]

		out_sock = socket(AF_INET, SOCK_STREAM)
		try:
		    out_sock.connect(destination[0], destination[1])
		except:
		    self.log("[ port %d -> %s ] Connection refused - restart client" % (localport, destination[0]))
		    exit(1)
		out_sock.send(packdict({dict_port : self.portlist[localport][2]}))

		clientcrypto = self.clientcrypto[destination[0]]
		daemoncrypto = self.daemoncrypto[destination[0]]

		if (fork() == 0):
		    self.log("[ port %d -> %s ] Connection established" % (localport, destination[0]))
		    data = ''
		    while 1:
			[iwtd, owtd, ewtd] = select([in_conn, out_sock],[],[])
			if in_conn in iwtd:
			    try:
				data = in_conn.recv(1024)
				if not data: break
				cipher = clientcrypto.encrypt(data)
				out_sock.send(cipher)
			    except error, detail:
				self.log("[ port %d -> %s ] Connection lost: %s" % (localport, destination[0], detail[1]))
				exit(1)
			if out_sock in iwtd:
			    try:
				data = out_sock.recv(1024)
				if not data: break
				plain = daemoncrypto.decrypt(data)
				in_conn.send(plain)
			    except error, detail:
				self.log("[ port %d -> %s ] Connection lost: %s" % (localport, destination[0], detail[1]))
				exit(1)
		    self.log("[ port %d -> %s ] Connection closed" % (localport, destination[0]))
		    in_conn.close()
		    out_sock.close()
		    execv('/bin/true', ())
		else:
		    in_conn.close()

parameter = Parameter()
parameter.parse()

signal.signal(signal.SIGINT, ctrlc_handler)
signal.signal(signal.SIGCHLD, sigchld_handler)

client = Client()
client.initialize(parameter.mappings, parameter.verbose, parameter.syslog)
client.listen()
